#include "Compiler.h"
#include "codeGen/Quadruple.h"
#include <filesystem>
#include <iostream>
#include <cassert>

extern std::filesystem::path binaryPath;
const std::string Compiler::extension = "s";
void Compiler::externalCommand(std::filesystem::path lat) {
	auto obj = lat;
	obj.replace_extension("o");
	auto cmd = std::string("as --32 '") + lat.string() + "' -o '" + obj.string() + "'";
	cerr << cmd << endl;
	system(cmd.data());
	auto lib = binaryPath.parent_path() / "lib";
	auto runtime = lib / "runtime.o";
	auto result = obj;
	result.replace_extension("");
	cmd = std::string("ld -m elf_i386 -L'") + lib.string() + "' '" + runtime.string() + "' '" + obj.string() + "' -o '" + result.string() + "'";
	cerr << cmd << endl;
	system(cmd.data());
}

std::string Compiler::compile(Visitable *v) {
	buf <<
		".globl _start\n"
	    ".globl exit\n"
		"_start:\n"
		"	pushl %ebp\n"
		"	movl %esp, %ebp\n"
		"	call main\n"
		"	pushl %eax\n"
		"	call exit\n";

	for (auto c : scope->classes) {
		c->calculateSize();
	}

	for (auto f : scope->functions) {
		compileFunction(f);
	}

	return buf.str();
}

void Compiler::compileFunction(FunctionInfoPtr f) {
	std::string name = mangleFunctionName(f);
	buf << ".globl \"" << name << "\"\n";

	if (!f->block) return;

	// TODO: jeśli funkcja jest klasowa, ustawić self na ... wartosc z ecx
	// c++ thiscall ecx, stack right-to-left (push c, push b, push a)

	buf << "\"" << name << "\":\n";
	scope->currentFunction = f;
	f->block->accept(this);

	int id = 1;
	for (auto v : vars) {
		if (v->constExpr && !v->info) {
			v->name = std::to_string(v->val);
			continue;
		}
		v->name = "@" + std::to_string(id);
		if (v->info) {
			v->name += "." + v->info->name + ":" + to_string(v->info->lineLocation);
		}
		id++;
	}

	int label = 1;
	for (auto q : quads) {
		if (auto l = dynamic_pointer_cast<QLabel>(q)) {
			if (!l->label.empty())
				l->label = to_string(label++) + "_" + l->label;
			else
				l->label = to_string(label++);
		}
	}
	for (auto q : quads) {
		buf << q->toString() << endl;
	}

	for (auto v : vars) {
		if (v->info)
			v->info->loc = nullptr;
	}

	quads.clear();
	scope->currentFunction = nullptr;
}

void Compiler::endBasicBlock() {
	for (auto v : vars) {
		if (v->info) v->info->loc = nullptr;
	}
}


/// expressions

void Compiler::visitEVar(EVar *p) {
	lastVar = alloc(p->pident_->var.lock());
}

void Compiler::visitELitInt(ELitInt *p) {
	lastVar = alloc(p->integer_);
}

void Compiler::visitELitTrue(ELitTrue *p) {
	lastVar = alloc(1);
}

void Compiler::visitELitFalse(ELitFalse *p) {
	lastVar = alloc(0);
}

void Compiler::visitEString(EString *p) {
	// TODO: zrobic liste stringow per app
	lastVar = alloc();
}

void Compiler::visitNullCast(NullCast *p) {
	lastVar = alloc(0);
}

// unary ops

void Compiler::visitNeg(Neg *p) {
	auto var = evalExpr(p->expr_);
	lastVar = alloc();
	addQuad<QAssign>(lastVar, Op::Neg, var);
}

void Compiler::visitNot(Not *p) {
	auto var = evalExpr(p->expr_);
	lastVar = alloc();
	addQuad<QAssign>(lastVar, Op::Not, var);
}

// binary ops

void Compiler::visitEMul(EMul *p) {
	auto l = evalExpr(p->expr_1);
	auto r = evalExpr(p->expr_2);
	lastVar = alloc();
	p->mulop_->accept(this);
	addQuad<QAssign>(lastVar, l, op, r);
}

void Compiler::visitEAdd(EAdd *p) {
	auto l = evalExpr(p->expr_1);
	auto r = evalExpr(p->expr_2);
	lastVar = alloc();
	p->addop_->accept(this);
	addQuad<QAssign>(lastVar, l, op, r);

}

void Compiler::visitERel(ERel *p) {
	auto l = evalExpr(p->expr_1);
	auto r = evalExpr(p->expr_2);
	lastVar = alloc();
	p->relop_->accept(this);
	addQuad<QAssign>(lastVar, l, op, r);
}

// lazy bools

void Compiler::visitEAnd(EAnd *p) {
	auto l = evalExpr(p->expr_1);
	auto labelLeft = make_shared<QLabel>("use_left");
	auto labelAfter = make_shared<QLabel>("end_and");

	addQuad<QJumpCond>(labelLeft, Op::Not, l);
	auto r = evalExpr(p->expr_2);
	lastVar = alloc();

	addQuad<QAssign>(lastVar, Op::Copy, r);
	addQuad<QJump>(labelAfter);
	addQuad(labelLeft);
	addQuad<QAssign>(lastVar, Op::Copy, l);
	addQuad(labelAfter);
}

void Compiler::visitEOr(EOr *p) {
	auto l = evalExpr(p->expr_1);
	auto labelLeft = make_shared<QLabel>("use_left");
	auto labelAfter = make_shared<QLabel>("end_or");

	addQuad<QJumpCond>(labelLeft, Op::Copy, l);

	auto r = evalExpr(p->expr_2);
	lastVar = alloc();

	addQuad<QAssign>(lastVar, Op::Copy, r);
	addQuad<QJump>(labelAfter);
	addQuad(labelLeft);
	addQuad<QAssign>(lastVar, Op::Copy, l);
	addQuad(labelAfter);
}

// complex extensions

void Compiler::visitEIndexAcc(EIndexAcc *p) {
	auto lhs = evalExpr(p->expr_1);
	auto index = evalExpr(p->expr_2);
	lastVar = alloc();
	auto type = dynamic_pointer_cast<Array>(lhs->type);
	if (!type) {
		throw ParseError("compiler evalExpr Variable has no Array type", p->expr_1);
	}
	if (!type->type_) {
		throw ParseError("compiler evalExpr Variable is Array of no type", p->expr_1);
	}
	if (type->type_->size() > 4) {
		// calculate quantifier, we multiply index because size is larger than 4
		auto q = alloc(type->type_->size() / 4);
		// calculate var = index * quantifier
		addQuad<QAssign>(lastVar, index, Op::Mul, q);

		// allocate new lastVar, index is now calculated multiplied index
		index = alloc();
		swap(index, lastVar);
	}
	auto quad = make_shared<QAccess>(lastVar, lhs, index, 4, 0);
	addQuad(quad);
}

void Compiler::visitEClsMmbr(EClsMmbr *p) {
	auto l = evalExpr(p->expr_);
	auto var = p->pident_->var.lock();
	assert(var);
	// it cannot be function, as function might only be used in EApp and are handled directly there
	assert(!var->toFunction());
	// Array length calculation works because of arrayLengthVar that has offset 0
	size_t offset = var->offset;
	if (dynamic_cast<EIndexAcc*>(p->expr_)) {
		// opt if EIndexAcc is used inside visitEClsMmbr
		auto access = dynamic_pointer_cast<QAccess>(quads.back());
		access->access.offset += offset;
		l->info = var;
		lastVar = l;
	} else {
		lastVar = alloc(var);
		auto quad = make_shared<QAccess>(lastVar, l, alloc(0), 0, offset);
		addQuad(quad);
	}
}

void Compiler::visitEApp(EApp *p) {
	auto funType = static_pointer_cast<Fun>(p->expr_->type);
	FunctionInfoPtr info;
	VariablePtr self;
	if (auto mem = dynamic_cast<EClsMmbr*>(p->expr_)) {
		// call tab[i].fun()
		// call obj.fun()
		self = evalExpr(mem->expr_);
		info = static_pointer_cast<FunctionInfo>(mem->pident_->var.lock());
	} else if (auto ident = dynamic_cast<EVar*>(p->expr_)) {
		auto klass = scope->currentFunction->klass.lock();
		info = static_pointer_cast<FunctionInfo>(ident->pident_->var.lock());
		auto otherClass = info->klass.lock();
		if (otherClass) {
			// call fun() = self.fun()
			assert(otherClass == klass);
			auto binding = scope->currentFunction->binding.lock();
			assert(binding);
			auto selfInfo = binding->variables.local("self");
			assert(selfInfo);
			self = alloc(selfInfo);
		} else {
			// call main()
		}
	} else {
		throw ParseError("Unimplemented EApp instantiation (neither EClsMmbr nor EVar", p);
	}

	int i = 1;
	for (auto param : *p->listexpr_) {
		auto var = evalExpr(param);
		addQuad<QParam>(var, i++);
	}

	if (self) {
		addQuad<QParam>(self, 0);
	}

	lastVar = alloc();
	addQuad<QCall>(lastVar, info, i, self);
}

void Compiler::visitENewArray(ENewArray *p) {
	// allocate enough memory with calloc, if it is class then init each entry...
	auto type = p->type_;
	size_t size = type->size();
	std::string virtSymbol;
	if (auto klass = dynamic_cast<ClassT*>(type)) {
		auto info = klass->pident_->var.lock()->toClass();
		size = info->size;
		virtSymbol = getVirtName(info);
	}
	auto count = evalExpr(p->expr_);
	lastVar = alloc();
	addQuad<QAlloc>(lastVar, size, virtSymbol, count);
}

void Compiler::visitENewClass(ENewClass *p) {
	// allocate enough memory with calloc, setup virt table pointer.
	auto info = p->pident_->var.lock()->toClass();
	size_t size = info->size;
	std::string virtSymbol = getVirtName(info);
	lastVar = alloc();
	addQuad<QAlloc>(lastVar, size, virtSymbol);
}

/// statements

void Compiler::visitInit(Init *p) {
	auto info = p->pident_->var.lock();
	assert(info);
	auto var = evalExpr(p->expr_);
	addQuad<QAssign>(alloc(info), Op::Copy, var);
}

void Compiler::assign(Expr* lval, VariablePtr val) {
	auto dest = evalLVal(lval);
	if (dest->info && dest->info->isInstanceVariable()) {
		// instance variable, need to write it to memory
		auto quad = dynamic_pointer_cast<QAccess>(quads.empty() ? nullptr : quads.back());
        assert(quad);
		auto loc = quad->access;
		quads.pop_back();
		addQuad<QWrite>(loc, val);
		return;
	} else {
		// local variable - only assign
		addQuad<QAssign>(dest, Op::Copy, val);
	}
}

void Compiler::visitAss(Ass *p) {
	auto val = evalExpr(p->expr_2);
	assign(p->expr_1, val);
}

void Compiler::visitIncr(Incr *p) {
	auto lhs = evalExpr(p->expr_);
	auto tmp = alloc();
	addQuad<QAssign>(tmp, lhs, Op::Plus, alloc(1));
	assign(p->expr_, tmp);
}

void Compiler::visitDecr(Decr *p) {
	auto lhs = evalExpr(p->expr_);
	auto tmp = alloc();
	addQuad<QAssign>(tmp, lhs, Op::Minus, alloc(1));
	assign(p->expr_, tmp);
}

void Compiler::visitRet(Ret *p) {
	auto var = evalExpr(p->expr_);
	endBasicBlock();
	addQuad<QReturn>(var);
}

void Compiler::visitVRet(VRet *p) {
	endBasicBlock();
	addQuad<QReturn>(nullptr);
}

void Compiler::visitCond(Cond *p) {
	auto var = evalExpr(p->expr_);
	auto after = make_shared<QLabel>("end_if");

	endBasicBlock(); // possible jump -> after
	addQuad<QJumpCond>(after, Op::Not, var);
	p->stmt_->accept(this);

	endBasicBlock(); // possible jump <- cond
	addQuad(after);
}

void Compiler::visitCondElse(CondElse *p) {
	auto elseBranch = make_shared<QLabel>("else");
	auto after = make_shared<QLabel>("end_else");

	auto var = evalExpr(p->expr_);

	endBasicBlock(); // possible jump -> else
	addQuad<QJumpCond>(elseBranch, Op::Not, var);

	p->stmt_1->accept(this);
	endBasicBlock(); // jump -> after
	addQuad<QJump>(after);

	addQuad(elseBranch);
	p->stmt_2->accept(this);

	endBasicBlock(); // jump <- cond
	addQuad(after);
}

void Compiler::visitWhile(While *p) {
	auto cond = make_shared<QLabel>("cond");

	endBasicBlock(); // jump <- loop -> cond
	addQuad<QJump>(cond);

	// jump <- loop
	auto loop = make_shared<QLabel>("loop");
	addQuad(loop);
	p->stmt_->accept(this);

	endBasicBlock(); // jump <- cond
	addQuad(cond);
	auto var = evalExpr(p->expr_);

	endBasicBlock(); // jump -> loop
	addQuad<QJumpCond>(loop, Op::Copy, var);
}

void Compiler::visitSExp(SExp *p) {
	evalExpr(p->expr_);
}

void Compiler::visitForEach(ForEach *p) {
	auto tab = evalExpr(p->expr_);
	throw ParseError("Unimplemented instruction!", p);
	// TODO: implement
}

