#ifndef COMPILER_HEADER
#define COMPILER_HEADER

#include "TypeDefs.h"
#include "Absyn.h"
#include "Skeleton.h"
#include <filesystem>
#include <utility>
#include "codeGen/Variable.h"
#include "ParseError.h"
#include "Info.h"
#include "codeGen/Quadruple.h"
#include "codeGen/BasicBlock.h"

using Op = Quadruple::Op;
class Compiler : public Skeleton
{
public:
	Compiler(const std::string& f, shared_ptr<Scope> s) : file(f), buf(), scope(std::move(s)) {};
	std::string compile(Visitable *v);

	static const std::string extension;
	static void externalCommand(std::filesystem::path file);

private:
	std::filesystem::path file;
	std::stringstream buf;
	shared_ptr<Scope> scope;
	vector<VariablePtr> vars;
	vector<QuadruplePtr> quads;
	Quadruple::Op op;

	VariablePtr lastVar;
	int lineno;

	void compileFunction(FunctionInfoPtr f);

	VariablePtr evalLVal(Visitable *expr) {
		auto info = evalExpr(expr)->info;
		if (!info) {
			throw ParseError("LValue expected", expr);
		}
		info->loc = nullptr;
		return alloc(info);
	}

	VariablePtr evalExpr(Visitable *expr) {
		if (!expr) throw runtime_error("No expr to eval");
		lineno = expr->lineno;
		lastVar = nullptr;
		try {
			expr->accept(this);
		} catch (const ParseError &err) {
			throw ParseError(err, expr);
		}
		if (!lastVar) throw ParseError("No variable found", expr);
		auto ret = lastVar;
		lastVar = nullptr;
		return ret;
	};

	std::string mangleFunctionName(FunctionInfoPtr f) const {
		if (auto c = f->klass.lock()) {
			return c->name + "::" + f->name;
		}
		return f->name;
	};

	std::string getVirtName(ClassInfoPtr c) const {
		return c->name + ":virt_table";
	};

	VariablePtr alloc(VarInfoPtr info) {
		if (info->loc && !info->isInstanceVariable()) {
			return info->loc;
		}
		auto v = make_shared<Variable>(info);
		vars.emplace_back(v);
		info->loc = v;
		return v;
	}
	template<typename... Args> VariablePtr alloc(Args... args) {
		auto v = make_shared<Variable>(args...);
		vars.emplace_back(v);
		return v;
	}

	template<typename T, typename... Args> QuadruplePtr addQuad(Args... args) {
		auto quad = make_shared<T>(args...);
		quad->lineno = lineno;
		quads.emplace_back(quad);
		return quad;
	}

	QuadruplePtr addQuad(QuadruplePtr quad) {
		quad->lineno = lineno;
		quads.emplace_back(quad);
		return quad;
	}

	void endBasicBlock();

	void assign(Expr* lval, VariablePtr val);

	void visitEVar(EVar *p) override;
	void visitEIndexAcc(EIndexAcc *p) override;
	void visitEClsMmbr(EClsMmbr *p) override;
	void visitEApp(EApp *p) override;
	void visitELitInt(ELitInt *p) override;
	void visitELitTrue(ELitTrue *p) override;
	void visitELitFalse(ELitFalse *p) override;
	void visitEString(EString *p) override;
	void visitENewArray(ENewArray *p) override;
	void visitENewClass(ENewClass *p) override;
	void visitNullCast(NullCast *p) override;
	void visitNeg(Neg *p) override;
	void visitNot(Not *p) override;
	void visitEMul(EMul *p) override;
	void visitEAdd(EAdd *p) override;
	void visitERel(ERel *p) override;
	void visitEAnd(EAnd *p) override;
	void visitEOr(EOr *p) override;



	void visitInit(Init *p) override;
	void visitAss(Ass *p) override;
	void visitIncr(Incr *p) override;
	void visitDecr(Decr *p) override;
	void visitRet(Ret *p) override;
	void visitVRet(VRet *p) override;
	void visitCond(Cond *p) override;
	void visitCondElse(CondElse *p) override;
	void visitWhile(While *p) override;
	void visitSExp(SExp *p) override;
	void visitForEach(ForEach *p) override;


	void visitPlus(Plus *p) override { op = Op::Plus; };
	void visitMinus(Minus *p) override { op = Op::Minus; };
	void visitTimes(Times *p) override { op = Op::Mul; };
	void visitDiv(Div *p) override { op = Op::Div; };
	void visitMod(Mod *p) override { op = Op::Mod; };
	void visitLTH(LTH *p) override { op = Op::LT; };
	void visitLE(LE *p) override { op = Op::LE; };
	void visitGTH(GTH *p) override { op = Op::GT; };
	void visitGE(GE *p) override { op = Op::GE; };
	void visitEQU(EQU *p) override { op = Op::EQ; };
	void visitNE(NE *p) override { op = Op::NEQ; };
};

#endif
